### gRL Link Assignment in Time-Space NW
### @author:Issei Yamano
### coding:SHIFT-JIS
### 2019/1/22

rm(list=ls(all=TRUE))
library(Matrix)
library(VGAM)
options(warn=-1) # Ignore all warnings
readdata.number <- -1 # -1 = all

################## Setting ######################

# Assumed Parameter  (TravelCost,AmusementStay,TimeDiscount)
assumed_b <- c(0.5, 2, 1) # Time Discount Beta = 1

# Time Step size
TS <- 10

# Input File area
input_area<-"C:/workspace_R/Yamano_Model/input"
input_project<-"virtual_9node_re"

## network
inputnodepass<-paste(input_area, input_project, "network/reduced_node.csv", sep = "/")
inputlinkpass<-paste(input_area, input_project, "network/reduced_link.csv", sep = "/")
inputmatrixpass<-paste(input_area, input_project, "network/reduced_matrix.csv", sep = "/")

## OD demand Data
inputdemandpass <- paste(input_area, input_project, "od_demand_2.csv", sep = "/")

# If you want to output the assignment result: 1,  else: 0
outputindex <- 1

# Directory for Output csv
outputdir <- paste("C:/workspace_R/MasterReserch/RL_LSM/output",input_project,"flow",sep="/")
setwd(outputdir)
options("scipen"=1000, "digits"=4)

############ RL Path Choice Problem #############

#########################################
########  reading & sampling    #########
#########################################

########## node data reading ############
node <- read.csv(inputnodepass)
nodeid <- sort(unique(node$NodeID))
N <- length(nodeid)
########## link data reading ############
link <- read.csv(inputlinkpass)
linkid <- sort(unique(link$LinkID))
L <- length(linkid)
########## link matrix reading ##########
matrix <- read.csv(inputmatrixpass)
names(matrix) <- c("k","a")
# Spatial connection matrix I
I <- array(0, dim = c(N, N))
for(i in 1:nrow(matrix)){
  kn <- matrix$k[i]
  an <- matrix$a[i]
  k <- (1:length(nodeid))[nodeid == kn]
  a <- (1:length(nodeid))[nodeid == an]
  I[k,a] <- 1
}

######### link connectiuon dummy ########
# Time Space Link connection
# input: ODlist
TSNW <- function(ODlist){
  Ilist <- list()
  # ԍ\邱Ƃő؍݂\
  for(i in 1:N){ I[i,i] <- 1 }
  
  # N_O̓B\ɊւwϐIt̒`
  Itlist <- array(0, dim = c(TS+1, N, N))
  II <- I
  Itlist[1,,] <- diag(N)
  for(ts in 2:(TS+1)){
    Itlist[ts,,] <- II
    II <- II%*%I
    II <- (II > 0)*1 + (II == 0)*0
  }

  # I_D̓B\ɊւwwϐItt̒`
  Ittlist <- array(0, dim = c(TS+1, N, N))
  for(ts in 1:(TS+1)){
    Ittlist[ts,,] <- t(Itlist[(TS+1)-ts+1,,])
  }

  OD <- nrow(ODlist)
  # wϐȂ玞ԓIڑs쐬
  for(od in 1:OD){
    
    # ODyAƂɎԓIڑs쐬
    Id <- array(0, dim = c(TS, N, N))
    on <- ODlist$O[od]
    dn <- ODlist$D[od]
    o <- (1:length(nodeid))[nodeid == on]
    d <- (1:length(nodeid))[nodeid == dn]
    for(ts in 1:TS){
      if(ts == 1){
        Id[ts,o,] <- I[o,]
        next
      }
      alist <- (1:N)[Itlist[ts+1,o,] == Ittlist[ts+1,d,]]
      for(a in alist){
        if(Itlist[ts+1,o,a] == Ittlist[ts+1,d,a]){
          klist <- (1:N)[I[,a]==1]
          for(k in klist){
            if( length((1:N)[Id[ts-1,,k] == 1]) != 0 ){
              Id[ts,k,a] <- 1
            }
          }
        } # if
      } # for a
    } # for ts
    
    Ilist[[od]] <- Id
    
    dn <- ODlist$D[od]
    dnn <- (1:length(nodeid))[nodeid == dn]
    
    # V0[TS, dnn, od] <- 0
    # z0[TS, dnn, od] <- 1
  } # for od
  
  return(Ilist)
}

############## Variables ################
# Node based
TT <- array(exp(0), dim = c(N, N))
Amusement <- array(exp(0), dim = c(N, N))
for(k in 1:N){
  kn <- node[k,]
  if( kn$Stay == 1 ){
    Amusement[k,k] <- exp(kn$Stay)
  }
  kn <- kn$NodeID
  for(a in 1:N){
    an <- nodeid[a]
    ka <- (1:L)[link$startNode == kn & link$endNode == an]
    if(length(ka)==0){
      TT[k,a] <- exp(0)
    }else{
      TT[k,a] <- exp(link$Length[ka])
    }
  }
}

#########################################
############ value function #############
#########################################
newV <- function(x) {
  # Setting
  theta <- x[length(x)]
  z <- array(1, dim = c(TS+1,N,OD)) #exp(Vd)

  for(od in 1:OD){
    # Calculate Utility matrix M
    M <- array(0, dim = c(TS,N,N))
    for(ts in 1:TS){
      Id <- Ilist[[od]]
      Mts <- Id[ts,,]*(TT^x[1])*(Amusement^x[2])
      M[ts,,] <- Mts
    }
    
    # Initialization (t=TS, V[ST,dnn,od]=0)
    # Calculate z(t)=Mz(t+1)+b from t=t-1 to t=0
    for(t in TS:1){
      zi <- M[t,,]%*%(z[t+1,,od]^theta)
      z[t,,od] <- (zi == 0)*1 + (zi != 0)*zi
    }
  } # for OD
  
  return(z)
}

#########################################
############## Probability ##############
#########################################
prob <- function(x, V) { 
  theta <- x[length(x)]
  Pt <- array(0, dim = c(OD,TS,N,N))
  
  for(od in 1:OD){
    for(ts in 1:TS){
      Id <- Ilist[[od]]
      z <- array(exp(theta*V[(ts+1),,od]), dim = c(N, 1))
      ZD <- array(rep(z), dim = c(N, N))
      ZD <- t(ZD)
      M <- Id[ts,,]*(TT^x[1])*(Amusement^x[2])
      Mz <- (M%*%z != 0)*(M%*%z) + (M%*%z == 0)
      MZ <- array(rep(Mz), dim = c(N,N))
      p <- (M*ZD)/MZ
      Pt[od,ts,,] <- p
    } # for TS
  } # for OD
  
  return(Pt)
}

#########################################
############## Assignment ###############
#########################################

########### Path Flow #############
assignment <- function(demand, p){
  # Generate Route data based on Choice Probability
  pathlist <- list()
  for(i in 1:OD){
    # For Each OD
    on <- demand$OriginNode[i]
    pod <- p[i,,,]
    
    # Create cumulative probability matrix
    for(ts in 1:TS){
      for(j in 1:N){
        if(j == 1){
          pod[ts,,j] <- pod[ts,,j]
        }else{
          pod[ts,,j] <- pod[ts,,(j-1)] + pod[ts,,j]
        }
      }
    }
    
    # Generate Route data as many as OD flow
    Q <- demand$Demand[i]
    if(Q == 0){
      pathlist[[i]] <- 0
      next
    }
    klist <- array(0, dim = c(Q,(TS+1)))
    for(q in 1:Q){
      rnd <- runif(TS)
      
      # Reserve First State
      kn <- on
      k <- (1:length(nodeid))[nodeid == kn]
      klist[q,1] <- kn
      
      for(ts in 1:TS){
        for(j in 1:N){
          if(rnd[ts] <= pod[ts,k,j]){
            k <- j
            kn <- nodeid[k]
            klist[q,(ts+1)] <- kn
            break
          }
        } # for j
      } # for ts
    } # for q
    pathlist[[i]] <- klist
  } # for od
  
  # Output
  pathlistcsv <- pathlist[[1]]
  if( length(pathlist) >= 2 ){
    for(i in 2:length(pathlist)){
      pathlistcsv <- rbind(pathlistcsv, pathlist[[i]])
    }
  }
  if(outputindex == 1){
    write.csv(pathlistcsv, "pathlist.csv", quote=FALSE, row.names=TRUE)
  }
  return(pathlist)
}

######### Estimation Data #########
dataset <- function(pathlist){
  # n: The number of People
  n <- 0
  for(od in 1:length(pathlist)){
    if(is.null(nrow(pathlist[[od]]))){
      next
    }else{
      n <- n + nrow(pathlist[[od]])
    }
  }
  
  # Create route data list for RL estimation
  Qlist <- list()
  route <- array(0, dim = c((n*TS),4))
  route <- data.frame(route)
  names(route) <- c("k","a","O","D")
  row <- 1
  for(od in 1:length(pathlist)){
    Qlist[[od]] <- nrow(pathlist[[od]])
    if(is.null(Qlist[[od]])){
      next
    }else{
      on <- unique(pathlist[[od]][,1])
      dn <- unique(pathlist[[od]][,(TS+1)])
      for(i in 1:nrow(pathlist[[od]]) ){
        for(ts in 1:TS){
          kn <- pathlist[[od]][i,ts]
          an <- pathlist[[od]][i,(ts+1)]
          route$k[row] <- kn
          route$a[row] <- an
          route$O[row] <- on
          route$D[row] <- dn
          row <- row + 1
        } # for ts
      } # for i
    } # if 
  } # for od
  
  # output
  if(outputindex == 1){
    write.csv(route, "route.csv", quote=FALSE, row.names=FALSE)
  }
  return(route)
}

########### Link Flow #############
# Stay and Link Flow by aggregating Route data (integer)
linkflow <- function(pathlist){
  fvlist <- array(0, dim = c(OD,L,TS))
  staylist <- array(0, dim = c(OD,N,TS))
  for(od in 1:OD){
    pathlistod <- pathlist[[od]]
    if(pathlistod != 0){
      for(i in 1:nrow(pathlist[[od]]) ){
        for(ts in 1:TS){
          kn <- pathlistod[i,ts]
          an <- pathlistod[i,ts+1]
          k <- (1:length(nodeid))[nodeid == kn]
          a <- (1:length(nodeid))[nodeid == an]
          if(kn == an){
            staylist[od,k,ts] <- staylist[od,k,ts] + 1
          }else{
            l <- (1:L)[link$startNode==kn & link$endNode==an]
            fvlist[od,l,ts] <- fvlist[od,l,ts] + 1
          }
        }
      }
    }
  }
  
  # summatation
  fv <- array(0, dim = c(L,TS))
  stay <- array(0, dim = c(N,TS))
  for(od in 1:OD){
    fv <- fv + fvlist[od,,]
    stay <- stay + staylist[od,,]
  }

  # output
  if(outputindex == 1){
    rownames(fv) <- as.character(linkid)
    rownames(stay) <- as.character(nodeid)
    write.csv(fv, "linkflow.csv", quote=FALSE, row.names=TRUE)
    write.csv(stay, "stayflow.csv", quote=FALSE, row.names=TRUE)
  }
  
  return( list(fvlist, staylist) )
}

########### Link Flow #############
# Stay and Link Flow based on probability (flow)
linkflow2 <- function(demand, p){
  flist <- array(0, dim = c(OD,TS,N,N))
  for(i in 1:OD){
    flow <- array(0, dim = c(TS,N,N))
    on <- demand$OriginNode[i]
    o <- (1:length(nodeid))[nodeid == on]
    g <- array(0, dim = c(N, 1))
    g[o] <- demand$Demand[i]
    for(ts in 1:TS){
      for(k in 1:N){
        flow[ts,k,] <- p[i,ts,k,]*g[k]
      }
      g <- colSums(flow[ts,,])
    }
    flist[i,,,] <- flow
  }
  
  fvlist <- array(0, dim = c(OD,L,TS))
  staylist <- array(0, dim = c(OD,N,TS))
  for(od in 1:OD){
    for(ts in 1:TS){
      fv <- array(0, dim = c(L,1))
      stay <- array(0, dim = c(N,1))
      for(i in 1:N){
        for(j in 1:N){
          if(flist[od,ts,i,j] != 0){
            if(i == j){
              stay[i] <- flist[od,ts,i,j]
            }else{
              kn <- nodeid[i]
              an <- nodeid[j]
              l <- (1:L)[link$startNode==kn & link$endNode==an]
              ln <- link$LinkID[l]
              fv[l] <- flist[od,ts,i,j]
            }
          }
        } # for j
      } # for i
      fvlist[od,,ts] <- fv
      staylist[od,,ts] <- stay
    } # for ts
  } # for od
  
  # summatation
  fv <- array(0, dim = c(L,TS))
  stay <- array(0, dim = c(N,TS))
  for(od in 1:OD){
    fv <- fv + fvlist[od,,]
    stay <- stay + staylist[od,,]
  }
  
  # output
  if(outputindex == 1){
    rownames(fv) <- as.character(linkid)
    rownames(stay) <- as.character(nodeid)
    write.csv(fv, "fv.csv", quote=FALSE, row.names=TRUE)
    write.csv(stay, "sv.csv", quote=FALSE, row.names=TRUE)
  }
  
  return( list(fvlist, staylist) )
}


# ------------------- Distribution (input: OD Flow and Parameters -------------------#

################### Assignment ####################
######## True OD demand reading #########
demand <- read.csv(inputdemandpass, nrows = readdata.number)
demandsize <- sum(demand$Demand)
ODlist <- demand[1:2]
names(ODlist) <- c("O","D")
OD <- nrow(ODlist)

####### Time-Space Connection Set #######
Ilist <- TSNW(ODlist)

############## Calculation ##############
b <- assumed_b
z <- newV(b)
z <- (z == 0)*1 + (z != 0)*z
V <- log(z)
p <- prob(b, V)
pathlist <- assignment(demand, p)
route <- dataset(pathlist)
LFlist <- linkflow(pathlist)
# LFlist <- linkflow2(demand, p)

fvlist <- LFlist[[1]]
staylist <- LFlist[[2]]

# Observed Dataset; Summatation of LF amd StayFlow
fv <- array(0, dim = c(L,TS))
stay <- array(0, dim = c(N,TS))
for(od in 1:OD){
  fv <- fv + fvlist[od,,]
  stay <- stay + staylist[od,,]
}
################### Assignment ####################


  


################# Dot Data Making #################
StartTime <- 0   # Set AM 08:00 to 0[s]
EndTime <- 600   # Set AM 08:10 to 600[s]
# if TS=10, time-interval is 360[s]=6[m]

# about TS
Stime <- round((EndTime - StartTime)/TS)
st <- array(0, dim=c((TS+1),1))
for(ts in 0:TS){
  st[(ts+1)] <- StartTime + ts*Stime
}

# Sensing Time Interval (unit: second)
Mtime <- 20
# Sensing Count
TM <- round((EndTime - StartTime)/Mtime)
# Sensing Time
mt <- array(0, dim=c((TM+1),2))
for(tm in 0:TM){
  mt[(tm+1),1] <- StartTime + tm*Mtime
  for(ts in 0:TS){
    if( mt[(tm+1),1] <= st[(ts+1)] ){
      mt[(tm+1),2] <- ts
      break
    }
  }
  mt[(tm+1),1] <- StartTime + tm*Mtime
}
mt <- data.frame(mt)
names(mt) <- c("Time","TS")


##### Speed Model #####
# Speed unit [100m/s]
speed <- 0.05
##### Speed Model #####

n_gpsdata <- (TM+1)*demandsize

mx <- array(0, dim=c(n_gpsdata,6))
j <- 0

##### Making Dot Data #####
for(od in 1:OD){
  for(n in 1:nrow(pathlist[[od]])){
    for(tm in 1:nrow(mt)){
      ts <- mt$TS[tm]
      tm2 <- (tm+j)
      if(ts==0){
        kn <- pathlist[[od]][n,1]
        kn <- node[node$NodeID == kn, ]
        mx[tm2,1] <- od
        mx[tm2,2] <- n
        mx[tm2,3] <- mt$Time[tm]
        mx[tm2,4] <- kn$Lat
        mx[tm2,5] <- kn$Lon
        mx[tm2,6] <- kn$NodeID
        next
      }
      
      kn <- pathlist[[od]][n,ts]
      kn <- node[node$NodeID == kn, ]
      an <- pathlist[[od]][n,(ts+1)]
      an <- node[node$NodeID == an, ]
      ka <- (1:L)[link$startNode == kn$NodeID & link$endNode == an$NodeID]
      
      if(length(ka)==0){ # if Stay
        mx[tm2,1] <- od
        mx[tm2,2] <- n
        mx[tm2,3] <- mt$Time[tm]
        mx[tm2,4] <- kn$Lat
        mx[tm2,5] <- kn$Lon
        mx[tm2,6] <- kn$NodeID
        next
      }
      
      ka <- link[ka,]
      # moveratio <- speed * (mt$Time[tm] - st[ts]) / ka$Length
      moveratio <- (mt$Time[tm] - st[ts]) / Stime
      if(moveratio >= 1){
        mx[tm2,1] <- od
        mx[tm2,2] <- n
        mx[tm2,3] <- mt$Time[tm]
        mx[tm2,4] <- an$Lat
        mx[tm2,5] <- an$Lon
        mx[tm2,6] <- an$NodeID
      }else{
        mx[tm2,1] <- od
        mx[tm2,2] <- n
        mx[tm2,3] <- mt$Time[tm]
        mx[tm2,4] <- (an$Lat - kn$Lat)*moveratio + kn$Lat
        mx[tm2,5] <- (an$Lon - kn$Lon)*moveratio + kn$Lon
        mx[tm2,6] <- ka$LinkID
      }
    } # for tm
    
    j <- j + nrow(mt)
    
  } # for n
} # for od

# Adding Random Positioning

#########################################
######### LonLat-Distance Model #########
#########################################
# Input: Distance r, Declination theta
LonLat <- function(rtheta_rand){
  # Store radius and argument
  r <- rtheta_rand[,1]
  theta <- rtheta_rand[,2]
  # Store the radius of the earth and the latitude of the target site
  R <- 6378137
  THETA <- 33.85/180 * pi
  # Calculate the distance per longitude 1 degree
  lon_1 <- 2*pi * R*cos(THETA) / 360
  # Calculate the distance per latitude 1 degree
  lat_1 <- 2*pi*R / 360
  # Calculate latitude and longitude
  # By dividing rcos(theta) and rsin(theta) by lon_1 and lat_1
  lon_error <- r*cos(theta) / lon_1
  lat_error <- r*sin(theta) / lat_1
  mxerror <- cbind(lon_error, lat_error)
  return(mxerror)
}

# Input: LonLat
Dis <- function(lon_error, lat_error){
  # n̔aƑΏە~n̈ܓx
  R <- 6378137
  THETA <- 33.85/180 * pi
  # ox1x̋vZĂ
  lon_1 <- 2*pi * R*cos(THETA) / 360
  # ܓx1x̋vZĂ
  lat_1 <- 2*pi*R / 360
  # lon_error, lat_errorɂꂼlon_1, lat_1|ĊevZ
  lon_dis <- lon_error * lon_1
  lat_dis <- lat_error * lat_1
  dis <- lon_dis^2 + lat_dis^2
  return(dis)
}

# Assume ditance follows Rayleigh distribution
r_rand <- array(rrayleigh( n_gpsdata, 5), dim=c(n_gpsdata,1)) # Rayleigh Distribution

# # Change Measurement error on Move or Stay
# # We can think about Spatial Measurement Error
# for(n in 1:nrow(pathlist[[od]])){
#   for(tm in 2:nrow(mt)){
#     tm2 <- (tm+(n-1)*31)
#     # # If Stay
#     # if(mx[tm2,6] == mx[(tm2-1),6] && mx[tm2,6] < 1000){
#     # 105,106,108,109Ȃ
#     if(mx[tm2,6] == 105 || mx[tm2,6] == 106 || mx[tm2,6] == 108 || mx[tm2,6] == 109|| mx[tm2,6] == 1013 || mx[tm2,6] == 1014 || mx[tm2,6] == 1017 || mx[tm2,6] == 1018  || mx[tm2,6] == 1019 || mx[tm2,6] == 1020 || mx[tm2,6] == 1023 || mx[tm2,6] == 1024 ){
#       r_rand[tm2] <- rrayleigh( 1, 15)
#     }
#   }
# }

theta_rand <- array(runif( n_gpsdata, 0, 2*pi), dim=c(n_gpsdata,1))
rtheta_rand <- cbind(r_rand, theta_rand)
mxrand2 <- LonLat(rtheta_rand)
mxdata <- mx[,4:5] + mxrand2


mx <- cbind(mx[,1:3], mxdata, mx[,6])
mx <- data.frame(mx)
names(mx) <- c("OD", "TripID", "Time", "Lat", "Lon", "True Position (Node/Link)")

# speed[km/h] calculation
mspeed <- array(0, dim=c(n_gpsdata,1))
for(i in 2:nrow(mx)){
  if(mx$TripID[i]==mx$TripID[i-1]){
    lon_error <- mx$Lon[i] - mx$Lon[i-1]
    lat_error <- mx$Lat[i] - mx$Lat[i-1]
    dis <- Dis(lon_error, lat_error)
    mspeed[i] <- (sqrt(dis)/1000) / ((mx$Time[i] - mx$Time[i-1])/3600)
  }
}

mx <- cbind(mx[,1:6], mspeed)
mx <- data.frame(mx)
names(mx) <- c("OD", "TripID", "Time", "Lat", "Lon", "True Position (Node/Link)", "Speed")

write.csv(mx[,-1], "dot_path.csv", quote=FALSE, row.names=FALSE)
##### Making Dot Data #####

################# Dot Data Making #################




